"""mtls (Mutual TLS) - A cli for creating short-lived client certiicates."""

import binascii
import os
import pkg_resources
import platform
import random
import re
import shutil
import subprocess
import sys
from configparser import ConfigParser
from pathlib import Path

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

import OpenSSL
import click
import gnupg
import json
import requests


class GPGDecryptionException(Exception):
    pass


class MutualTLS:
    MISSING_CONFIGURATION = """
    Configuration missing for mtls at {}/{}
    """
    MISSING_CONFIGURATION_FOR_SERVER = """
    Configuration missing for server: {server}

    Please ensure that you have a configuration for you server similar to:
    [{server}]
    email=foo@example.com
    url=https://ca.example.com

    For more details see config.ini.example
    """

    def __init__(self, server, options={}):
        self.HOME = os.environ.get("HOME")
        self.GNUPGHOME = os.environ.get("GNUPGHOME")
        self.USER = os.environ.get("USER")
        self.options = options
        self.override = False
        if options["config"] is None:
            self.CONFIG_FOLDER_PATH = "{}/.config/mtls".format(self.HOME)
            self.CONFIG_FILE = "config.ini"
            self.config_file_path = "{config_path}/{config_file}".format(
                config_path=self.CONFIG_FOLDER_PATH,
                config_file=self.CONFIG_FILE,
            )
        else:
            self.CONFIG_FOLDER_PATH = "/".join(
                options["config"].split("/")[:-1]
            )
            self.CONFIG_FILE = options["config"].split("/")[-1]
            self.config_file_path = options["config"]
        self.USER_KEY = "{}.key.gpg".format(self.USER)
        self.GNUPGHOME = os.environ.get("GNUPGHOME")
        if self.GNUPGHOME is None:
            self.GNUPGHOME = "{}/{}".format(os.environ.get("HOME"), ".gnupg")
        self.gpg = gnupg.GPG(gnupghome=self.GNUPGHOME)
        self.gpg.encoding = "utf-8"
        self.config = self.get_config()
        self.server = server
        self._make_server_dir_if_missing()
        self.server_in_config()
        self.openssl_format = serialization.PrivateFormat.TraditionalOpenSSL
        self.no_encyption = serialization.NoEncryption()
        self.friendly_name = "{org} - {user}@{hostname}".format(
            org=self.config.get(self.server, "organization_name"),
            user=str(os.environ.get("USER")),
            hostname=str(platform.uname()[1]),
        )
        self.BASE_SERVER_PATH = "{base_path}/{server}".format(
            base_path=self.CONFIG_FOLDER_PATH, server=self.server
        )
        self.pfx_path = "{base}/{server}.pfx".format(
            base=self.BASE_SERVER_PATH, server=self.server
        )
        self.cert_file_path = "{base}/{server}.pem".format(
            base=self.BASE_SERVER_PATH, server=self.server
        )
        self.ca_cert_file_path = "{base}/{server}_Root_CA.pem".format(
            base=self.BASE_SERVER_PATH, server=self.server
        )
        self.crl_file_path = "{base}/crl.pem".format(
            base=self.BASE_SERVER_PATH, server=self.server
        )

    def check_revoked(self, cert):
        with open(self.crl_file_path, "rb") as f:
            crl = x509.load_pem_x509_crl(
                data=f.read(), backend=default_backend()
            )
            if cert.issuer != crl.issuer:
                click.secho("Cert does not match CRL", fg="red")
                sys.exit(1)
            for revoked in crl:
                if cert.serial == revoked.serial_number:
                    return True

            return False

    def create_cert(self, output):
        if output:
            self.override = True
        self._create_db()
        cert = None
        if not self._has_root_cert():
            self._get_and_set_root_cert()
        if sys.platform == "darwin":
            (valid, exists, revoked) = self.check_valid_cert(
                name=self.cert_file_path
            )
        else:
            (valid, exists, revoked) = self.check_valid_cert(
                name=self.friendly_name
            )
        if valid is True:
            click.secho("Reusing valid certificate", fg="green")
            sys.exit(0)
        if valid is False and exists is True:
            if sys.platform == "darwin":
                self.delete_cert_by_name(self.friendly_name)
            else:
                self.delete_cert_by_name(self.friendly_name)
        if valid is False and revoked is False and exists is True:
            cert = self.get_cert_from_file()
        csr = self.get_csr()
        key = self.get_key_or_generate()
        if csr is None:
            csr = self.generate_csr(key)
        else:
            click.secho(
                "Reusing previously generated CSR for {server}".format(
                    server=self.server
                ),
                fg="green",
            )
        cert_str = self.sign_and_send_to_server(csr)
        if cert_str is None:
            click.echo("Could not retrieve certificate from server")
            sys.exit(1)
        cert = self.convert_to_cert(cert_str)
        try:
            with open(self.cert_file_path, "w") as cert_file:
                click.echo("Writing file to {}".format(self.cert_file_path))
                cert_file.write(
                    cert.public_bytes(serialization.Encoding.PEM).decode(
                        "utf-8"
                    )
                )
        except Exception as e:
            click.secho(
                "Could not write certificate to {}".format(
                    self.cert_file_path
                ),
                fg="red",
            )
        if cert is None:
            click.echo("Could not convert to certificate")
            sys.exit(1)
        p12 = OpenSSL.crypto.PKCS12()
        pkey = OpenSSL.crypto.PKey.from_cryptography_key(key)
        fpbytes = cert.fingerprint(hashes.SHA1())
        fp = binascii.hexlify(fpbytes)
        if not self.override:
            self.update_config_value(
                "current_sha", fp.decode("UTF-8"), self.server
            )
        certificate = OpenSSL.crypto.X509.from_cryptography(cert)
        p12.set_privatekey(pkey)
        p12.set_certificate(certificate)
        p12.set_friendlyname(bytes(self.friendly_name, "UTF-8"))
        pwd = self._genPW()
        if output:
            self.pfx_path = output
            pw = self.encrypt(pwd, self.config.get(self.server, "fingerprint"))
            pfx_base = "/".join(self.pfx_path.split("/")[:-1])
            pw_file = self.pfx_path.split("/")[-1].split(".")[:-1]
            pw_file += ".password.asc"
            pw_file = "".join(pw_file)
            pw_file = "{}/{}".format(pfx_base, pw_file)
            with open(pw_file, "wb") as pwfile:
                click.echo("Writing password to: {}".format(pw_file))
                pwfile.write(str(pw).encode("UTF-8"))
        with open(self.pfx_path, "wb") as f:
            f.write(p12.export(passphrase=bytes(pwd, "UTF-8")))
        if not output:
            self.update_cert_storage(self.pfx_path, pwd)
            self._chrome_notice()
            self._firefox_notice()

    def _chrome_notice(self):
        CHROME_NOTICE = (
            "If using Chrome/Chromium you may have to restart before these "
            + "certificates take effect. chrome://restart"
        )
        click.secho(CHROME_NOTICE, fg="green")

    def _firefox_notice(self):
        click.secho(
            "If using Firefox you may have to restart "
            + "before these certificates take effect. about:profiles",
            fg="green",
        )

    def _has_root_cert(self):
        if sys.platform == "darwin":
            name = self.ca_cert_file_path
        else:
            name = "{server} Root CA".format(server=self.server)
        if self.check_valid_cert(name, usage="CT,T,T", is_root=True):
            return True
        return False

    def _get_and_set_root_cert(self):
        response = self.send_request(
            server_url="{url}/ca".format(
                url=self.config.get(self.server, "url")
            ),
            method="get",
        )
        try:
            data = response.json()
        except Exception:
            click.secho(
                "Error parsing Root Certificate from server.", fg="red"
            )
            sys.exit(1)
        # Update the issuer name directly from the server into your config
        self.config.set(self.server, "issuer", data["issuer"])
        self.update_config()
        # Write the file to the CA Cert File path so that it's accessible to
        # the user and subsequent calls later.
        with open(self.ca_cert_file_path, "w") as ca_cert:
            ca_cert.write(data["cert"])
        self.add_root_ca_to_store(self.ca_cert_file_path)

    def add_root_ca_to_store(self, ca_cert_file_path):
        click.echo("Adding root certificate to certificate store...")
        paths = self._get_certdb_paths()
        if sys.platform == "darwin":
            cmds = []
            add_trust_keychain = [
                "security",
                "add-trusted-cert",
                "-p",
                "ssl",
                "-r",
                "trustAsRoot",
                ca_cert_file_path,
            ]
            import_keychain = ["security", "import", ca_cert_file_path]
            cmds = [add_trust_keychain, import_keychain]
            for cmd in cmds:
                try:
                    output = self._run_cmd(cmd, capture_output=True)
                except Exception as e:
                    click.echo("Error")
                    click.echo(e)
            # Override paths for darwin to only handle firefox
            paths = self._firefox_certdb_location()
        if sys.platform == "linux" or sys.platform == "linux2" or "darwin":
            for path in paths:
                cmd = [
                    "certutil",
                    "-A",
                    "-d",
                    path,
                    "-t",
                    "CT,CT,CT",
                    "-i",
                    ca_cert_file_path,
                    "-n",
                    "{server} Root CA".format(server=self.server),
                ]
                try:
                    output = self._run_cmd(cmd, capture_output=True)
                except Exception as e:
                    click.echo("Error")
                    click.echo(e)

    def delete_cert_by_name(self, name):
        paths = self._get_certdb_paths()
        if sys.platform == "darwin":
            fingerprint = self.config.get(
                self.server, "current_sha", fallback=""
            )
            if fingerprint != "":
                click.secho(
                    "Deleting invalid/expired certificates for {}".format(
                        fingerprint
                    ),
                    fg="yellow",
                )
                delete_identity_cmd = [
                    "security",
                    "delete-identity",
                    "-Z",
                    fingerprint,
                ]
                output = self._run_cmd(
                    delete_identity_cmd, capture_output=True
                )
            # Override path to just be firefox on darwin for the next command
            paths = self._firefox_certdb_location()
        if sys.platform in ["linux", "linux2", "darwin"]:
            click.secho(
                "Deleting invalid/expired certificates for {}".format(name),
                fg="yellow",
            )
            for path in paths:
                cmd = ["certutil", "-D", "-d", path, "-n", name]

                try:
                    output = self._run_cmd(cmd, capture_output=True)
                except Exception as e:
                    click.echo(
                        "Error deleting certificate with name: {}".format(name)
                    )
                    click.echo(e)

    def check_valid_cert(self, name=None, usage="V", is_root=False):
        if name is None:
            click.echo("A valid certificate name is required")
            sys.exit(1)
        paths = self._get_certdb_paths()
        is_valid = False
        revoked = False
        cert_exists = os.path.isfile(self.cert_file_path)
        if not cert_exists:
            return is_valid, cert_exists, revoked
        if self.config.get(self.server, "issuer", fallback=None) is None:
            # If the config doesn't have an issuer, we can by default know that
            # a user has not received this information from the server or set
            # it themselves and can assume they don't have a certificate yet
            return False
        if sys.platform == "darwin":
            # Name is the path to the certificate file, because
            # security does not support verification once a certificate is
            # installed, so we must check that:
            # a) the certificate file exist
            # b) is it actually valid
            # We are only checking Keychain as we cannot verify that Firefox is
            # actually installed and we should not be making that assumption.
            if is_root:
                cmd = [
                    "security",
                    "find-certificate",
                    "-c",
                    self.config.get(self.server, "issuer"),
                ]
                find_cert_output = self._run_cmd(cmd, capture_output=True)
                if "The specified item could not be found" in str(
                    find_cert_output.stderr, "UTF-8"
                ):
                    return is_valid, cert_exists, revoked
                cmd = ["security", "verify-cert", "-r", self.ca_cert_file_path]
            else:
                find_cert_output = self._run_cmd(
                    ["security", "find-identity", "-p", "ssl-client", "-v"],
                    capture_output=True,
                )
                stdout_str = find_cert_output.stdout.decode("UTF-8")
                if self.friendly_name not in stdout_str:
                    return is_valid, cert_exists, revoked
                if "The specified item could not be found" in str(
                    find_cert_output.stderr, "UTF-8"
                ):
                    return is_valid, cert_exists, revoked
                cmd = [
                    "security",
                    "verify-cert",
                    "-c",
                    name,
                    "-r",
                    self.ca_cert_file_path,
                ]

            output = self._run_cmd(cmd, capture_output=True)
            if "CSSMERR_TP_NOT_TRUSTED" in output.stdout.decode("UTF-8"):
                is_valid = False
            if "CSSMERR_TP_CERT_EXPIRED" in output.stderr.decode("UTF-8"):
                is_valid = False
        elif sys.platform == "linux" or sys.platform == "linux2":
            for path in paths:
                cmd = [
                    "certutil",
                    "-V",
                    "-u",
                    usage,
                    "-d",
                    path,
                    "-n",
                    "{name}".format(name=name),
                ]
                output = self._run_cmd(cmd, capture_output=True)
                if "could not find certificate" in str(output.stderr, "UTF-8"):
                    # No certificate found so it's not valid, exists or revoked
                    is_valid = False
                    revoked = False
                if "certificate is invalid" in str(output.stdout, "UTF-8"):
                    is_valid = False
                    revoked = self.check_revoked(self.get_cert_from_file())
                if "validation failed" in str(output.stderr, "UTF-8"):
                    is_valid = False
                    revoked = self.check_revoked(self.get_cert_from_file())
        return is_valid, cert_exists, revoked

    def get_cert_from_file(self):
        with open(self.cert_file_path, "rb") as cert_file:
            return x509.load_pem_x509_certificate(
                cert_file.read(), default_backend()
            )

    def decrypt(self, data, is_file=True):
        if is_file:
            data = self.gpg.decrypt_file(data)
        else:
            data = self.gpg.decrypt(data)

        if data.ok is False:
            raise GPGDecryptionException(data.status)

        return str(data)

    def get_csr(self):
        csr_path = "{}/{}.csr.asc".format(self.CONFIG_FOLDER_PATH, self.server)
        if not os.path.isfile(csr_path) or self.override:
            return None
        click.echo("Decrypting CSR...")
        try:
            csr_str = self.decrypt(open(csr_path, "rb"))
        except GPGDecryptionException:
            click.secho("Failed to decrypt CSR, invalid password.", fg="red")
            sys.exit(1)
        return x509.load_pem_x509_csr(
            bytes(csr_str, "utf-8"), default_backend()
        )

    def _genPW(self):
        wordList = []
        with pkg_resources.resource_stream(
            __name__, "share/password_word_list"
        ) as wordFile:
            for line in wordFile:
                wordList.append(line.decode().rstrip("\n"))
        pw = []
        for c in range(10):
            pw.append(random.choice(wordList))
        return " ".join(pw).rstrip()

    def convert_to_cert(self, cert):
        try:
            cert = bytes(str(cert), "utf-8")
            cert = x509.load_pem_x509_certificate(
                cert, backend=default_backend()
            )
            return cert
        except Exception as e:
            click.secho(
                "Failure to load PEM x509 Certificate: {}".format(e),
                fg="red",
                err=True,
            )

    def _run_cmd(self, args, capture_output=False):
        if capture_output:
            return subprocess.run(
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
        return subprocess.run(args)

    def update_cert_storage(self, cert_file_path, cert_pw):
        if sys.platform == "linux" or sys.platform == "linux2":
            paths = self._get_certdb_paths()
            for path in paths:
                try:
                    cmd = [
                        "pk12util",
                        "-i",
                        cert_file_path,
                        "-d",
                        path,
                        "-W",
                        cert_pw,
                        "-n",
                        str(platform.uname()[1]),
                    ]
                    output = self._run_cmd(cmd, capture_output=True)
                    if "SEC_ERROR_REUSED_ISSUER_AND_SERIAL" in str(
                        output.stderr, "UTF-8"
                    ):
                        self.delete_cert_by_name(self.friendly_name)
                        self.update_cert_storage(cert_file_path, cert_pw)
                        return
                except Exception as e:
                    cse = "Could not add certificate to certificate store"
                    click.secho(cse, fg="red", err=True)
                    click.echo(e)
        elif sys.platform == "darwin":
            try:
                # Add to keychain
                self._run_cmd(
                    [
                        "security",
                        "import",
                        cert_file_path,
                        "-f",
                        "pkcs12",
                        "-x",
                        "-P",
                        cert_pw,
                    ],
                    capture_output=True,
                )
            except Exception as e:
                click.echo("Could not add certificate to certificate store")
            try:
                # Add to FireFox Store
                paths = self._firefox_certdb_location()
                for path in paths:
                    cmd = [
                        "pk12util",
                        "-i",
                        cert_file_path,
                        "-d",
                        path,
                        "-W",
                        cert_pw,
                        "-n",
                        str(platform.uname()[1]),
                    ]
                    self._run_cmd(cmd, capture_output=True)
            except Exception as e:
                click.secho(
                    "Could not add certificate to certificate store",
                    fg="red",
                    err=True,
                )
                click.echo(e)
        else:
            try:
                self._run_cmd(
                    ["certutil.exe", "-viewstore", "-user", "root"],
                    capture_output=True,
                )
            except Exception as e:
                click.echo("Could not add certificate to certificate store")
                click.echo(e)

    def _make_server_dir_if_missing(self):
        path = "{config}/{server}".format(
            config=self.CONFIG_FOLDER_PATH, server=self.server
        )
        if not os.path.isdir(path):
            os.makedirs(path)

    def _create_db(self):
        path = self._primary_certdb_location()
        if sys.platform == "linux" or sys.platform == "linux2":
            if not os.path.isdir(path):
                os.makedirs(path)
                # This assumes that you're using an encrypted hard drive
                # Othewise we need to look into a way to allow user input
                # Probably swapping subprocess.call with subprocess.popen so
                # that users can manually enter passwords
                click.secho("Making nssdb at {}".format(path), fg="green")
                subprocess.call(
                    ["certutil", "-d", path, "-N", "--empty-password"]
                )

    def _get_certdb_paths(self):
        paths = [
            self._primary_certdb_location()
        ] + self._firefox_certdb_location()
        return paths

    def _primary_certdb_location(self):
        return os.path.join(self.HOME, ".pki/nssdb")

    def _firefox_certdb_location(self):
        base_path = None
        paths = []
        if sys.platform == "linux" or sys.platform == "linux2":
            base_path = os.path.join(self.HOME, ".mozilla/firefox")
        elif sys.platform == "darwin":
            # Make directory if it doesn't exist
            base_path = os.path.join(
                self.HOME, "Library/Application Support/Firefox/Profiles/"
            )
        elif sys.platform == "win32":
            base_path = os.path.join(
                os.getenv("USERPROFILE"), "\\AppData\\Local\\Firefox"
            )
        if base_path is not None:
            posix_paths = list(Path(base_path).rglob("cert*.db"))
            for pp in posix_paths:
                paths.append("/".join(str(pp).split("/")[:-1]))
        return paths

    def check_for_config(self):
        """Check if the config exists, otherwise exit."""
        config_dir_exists = os.path.isdir(self.CONFIG_FOLDER_PATH)
        config_exists = os.path.isfile(self.config_file_path)

        if not config_dir_exists or not config_exists:
            msg = self.MISSING_CONFIGURATION.format(
                self.CONFIG_FOLDER_PATH, self.CONFIG_FILE
            )
            click.echo(msg)
            sys.exit(1)

    def get_config(self):
        """Gets config from file.

        Returns:
            config
        """
        self.check_for_config()
        config = ConfigParser()
        config.read(self.config_file_path)
        return config

    def update_config_value(self, key, value, namespace="DEFAULT"):
        self.config.set(namespace, key, value)
        self.update_config(show_msg=False)

    def update_config(self, show_msg=True):
        if show_msg:
            click.secho(
                "Updating config file settings for {server}".format(
                    server=self.server
                ),
                fg="green",
            )
        with open(self.config_file_path, "w") as config_file:
            self.config.write(config_file)

    def server_in_config(self):
        """Determines if the set server is in the config, otherwise exit."""
        if self.server is None and len(self.config.sections()) > 1:
            click.echo(
                "You have multiple servers configured, please "
                + "selection one with the --server (-s) option"
            )
        if self.server not in self.config:
            click.echo(
                self.MISSING_CONFIGURATION_FOR_SERVER.format(
                    server=self.server
                )
            )
            sys.exit(1)

    def encrypt(self, data, recipient, sign=False):
        """Encrypt data using PGP to recipient."""
        if sign is True:
            click.echo("Encrypting and Signing data...")
        return self.gpg.encrypt(data, recipient, sign=sign)

    def get_key_or_generate(self):
        """Get users key from file or generate a new RSA Key.

        Returns:
            key - RSA Key
        """
        key = None
        key_path = "{config_folder}/{user_key}".format(
            config_folder=self.CONFIG_FOLDER_PATH, user_key=self.USER_KEY
        )
        if os.path.isfile(key_path):
            click.echo("Decrypting User Key...")
            try:
                key_data = self.decrypt(open(key_path, "rb"))
            except GPGDecryptionException:
                click.secho(
                    "Failed to decrypt user key. Invalid password", fg="red"
                )
                sys.exit(1)
            byte_key_data = bytes(str(key_data), "utf-8")
            key = serialization.load_pem_private_key(
                byte_key_data, password=None, backend=default_backend()
            )
        else:
            click.echo("Generating User Key")
            key = rsa.generate_private_key(
                public_exponent=65537, key_size=4096, backend=default_backend()
            )
            key_data = key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=self.openssl_format,
                encryption_algorithm=self.no_encyption,
            )
            user_fingerprint = self.config.get(self.server, "fingerprint")
            encrypted_key = self.encrypt(key_data, user_fingerprint)
            user_email = self.config.get(self.server, "email")
            click.echo("Encrypting file to {}".format(user_email))
            with open(key_path, "w") as f:
                f.write(str(encrypted_key))
        return key

    def generate_csr(self, key):
        """Generates a CSR.

        Args:
            key - The users key

        Returns:
            csr - The CSR
        """
        click.secho(
            "Generating CSR for {server}".format(server=self.server),
            fg="yellow",
        )
        country = self.config.get(self.server, "country", fallback=None)
        state = self.config.get(self.server, "state", fallback=None)
        locality = self.config.get(self.server, "locality", fallback=None)
        organization_name = self.config.get(self.server, "organization_name")
        email = self.config.get(self.server, "email")
        csr_subject_arr = [
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
            x509.NameAttribute(NameOID.COMMON_NAME, self.friendly_name),
            x509.NameAttribute(NameOID.EMAIL_ADDRESS, email),
        ]
        if state:
            csr_subject_arr.append(
                x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state)
            )
        if country:
            csr_subject_arr.append(
                x509.NameAttribute(NameOID.COUNTRY_NAME, country)
            )
        if locality:
            csr_subject_arr.append(
                x509.NameAttribute(NameOID.LOCALITY_NAME, locality)
            )
        csr = (
            x509.CertificateSigningRequestBuilder()
            .subject_name(x509.Name(csr_subject_arr))
            .sign(key, hashes.SHA256(), default_backend())
        )
        csr_fname = "{}.csr.asc".format(self.server)

        # If the user overrides the ini configuration programatically, don't
        # save the CSR as this could be a one off.
        if self.override:
            return csr

        with open(
            "{config}/{server}/{csr}".format(
                config=self.CONFIG_FOLDER_PATH,
                server=self.server,
                csr=csr_fname,
            ),
            "wb",
        ) as f:
            enc_csr = self.encrypt(
                csr.public_bytes(serialization.Encoding.PEM).decode("utf-8"),
                self.config.get(self.server, "fingerprint"),
            )
            f.write(bytes(str(enc_csr), "utf-8"))
        return csr

    def _get_path(self, path):
        """Gets the absolute path given a path."""
        if not os.path.isabs(path):
            return os.path.abspath(os.path.join(self.CONFIG_FOLDER_PATH, path))

    def send_request(self, payload=None, method="post", server_url=None):
        if server_url is None:
            server_url = self.config.get(self.server, "url")
        if method == "post":
            if payload is None:
                click.secho(
                    "Payload missing for request. Cancelling",
                    fg="red",
                    err=True,
                )
            return requests.post(server_url, json=payload, verify=True)
        if method == "delete":
            if payload is None:
                click.secho(
                    "Payload missing for request. Cancelling",
                    fg="red",
                    err=True,
                )
            return requests.delete(server_url, json=payload, verify=True)
        if method == "get":
            return requests.get(server_url, verify=True)
        click.secho(
            "Failed to properly send request to server, "
            + "invalid method passed to MutualTLS.send_request"
        )
        sys.exit(1)

    def gen_sig(self, data, echo_msg):
        click.echo(echo_msg)
        if self.options["gpg_password"]:
            return self.gpg.sign(
                data,
                keyid=self.config.get(self.server, "fingerprint"),
                passphrase=self.options["gpg_password"],
                detach=True,
                clearsign=True,
            )
        else:
            return self.gpg.sign(
                data,
                keyid=self.config.get(self.server, "fingerprint"),
                detach=True,
                clearsign=True,
            )

    def sign_and_send_to_server(self, csr):
        """Sign and send to server.

        Args:
            csr - The CSR

        Returns:
            cert - the certificate
        """
        csr_public_bytes = csr.public_bytes(serialization.Encoding.PEM)
        msg = "Signing CSR for verification on server..."
        signature = self.gen_sig(csr_public_bytes, msg)
        payload = {
            "csr": csr_public_bytes.decode("utf-8"),
            "signature": str(signature),
            "lifetime": self.config.get(
                self.server, "lifetime", fallback=64800
            ),
            "type": "CERTIFICATE",
        }
        response = self.send_request(payload)
        try:
            response = response.json()
        except Exception as e:
            click.secho(
                "Error handling response from server. Bailing", fg="red"
            )
            print(response.text)
            sys.exit(-1)
        if response.get("error", False):
            click.echo(response.get("msg"))
            sys.exit(1)
        return str(response["cert"])

    def revoke_cert(self, fingerprint, serial_number, common_name):
        payload = {"query": {}, "type": "CERTIFICATE"}
        if fingerprint is not None:
            payload["query"]["fingerprint"] = fingerprint
        if serial_number is not None:
            payload["query"]["serial_number"] = serial_number
        if common_name is not None:
            payload["query"]["common_name"] = common_name
        msg = "Signing Revoke Request..."
        payload["signature"] = str(
            self.gen_sig(json.dumps(payload["query"]).encode("UTF-8"), msg)
        )
        response = self.send_request(payload, method="delete")
        response = response.json()
        if response.get("error", False):
            click.echo(response.get("msg"))
            sys.exit(1)
        click.echo("Certificate Revoked")

    def add_user(self, fingerprint, is_admin=False):
        msg = "Signing Add User Request..."
        payload = {
            "type": "ADMIN" if is_admin else "USER",
            "fingerprint": fingerprint,
            "signature": str(self.gen_sig(fingerprint.encode("UTF-8"), msg)),
        }
        response = self.send_request(payload, method="post")
        response = response.json()
        if response.get("error", False):
            click.echo(response.get("msg"))
            sys.exit(1)
        if is_admin:
            _type = "Admin"
        else:
            _type = "User"
        click.secho(
            "Added {_type}: {fingerprint}".format(
                fingerprint=fingerprint, _type=_type
            ),
            fg="green",
        )

    def remove_user(self, fingerprint, is_admin=False):
        msg = "Signing Remove User Request..."
        payload = {
            "type": "ADMIN" if is_admin else "USER",
            "fingerprint": fingerprint,
            "signature": str(self.gen_sig(fingerprint.encode("UTF-8"), msg)),
        }
        response = self.send_request(payload, method="post")
        response = response.json()
        if response.get("error", False):
            click.echo(response.get("msg"))
            sys.exit(1)
        if is_admin:
            _type = "Admin"
        else:
            _type = "User"
        click.secho(
            "Removed {_type}: {fingerprint}".format(
                fingerprint=fingerprint, _type=_type
            ),
            fg="green",
        )

    def set_user_options(self, options):
        for key in options:
            self.config.set(self.server, key, options.get(key))

        if options.get("common_name"):
            self.friendly_name = "{org} - {name}".format(
                org=self.config.get(self.server, "organization_name"),
                name=options["common_name"],
            )

        if options.get("friendly_name"):
            self.friendly_name = "{org} - {name}".format(
                org=self.config.get(self.server, "organization_name"),
                name=options["friendly_name"],
            )

    def get_crl(self, output):
        if not output:
            click.echo("Retrieving CRL from server...")
        response = self.send_request(
            server_url=self.config.get(self.server, "url") + "/crl",
            method="get",
        )
        if response.status_code != 200:
            click.secho(
                "Failed to retrieve CRL from {}".format(self.server),
                fg="red",
                err=True,
            )
        if output:
            click.echo(response.text)
        else:
            click.echo("Writing CRL to {}".format(self.crl_file_path))
            with open(self.crl_file_path, "wb") as crl_file:
                crl_file.write(bytes(response.text, "UTF-8"))



import logging
import platform
import os
import random
import re
import subprocess
import time
import traceback
import unittest

from click.testing import CliRunner
from configparser import ConfigParser
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.backends import openssl
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
import click
import docker
import gnupg
import tempfile
import requests

from mtls.cli import cli


def getListOfFiles(dirName):
    listOfFile = os.listdir(dirName)
    allFiles = list()
    for entry in listOfFile:
        fullPath = os.path.join(dirName, entry)
        if os.path.isdir(fullPath):
            allFiles = allFiles + getListOfFiles(fullPath)
        else:
            allFiles.append(fullPath)
    return allFiles


logging.disable(logging.CRITICAL)
MTLS_SERVER_VERSION = os.environ.get("MTLS_SERVER_VERSION") or "v0.16.3"
MTLS_IMAGE = os.environ.get("MTLS_IMAGE") or "drgrove/mtls-server"


def generate_key():
    return rsa.generate_private_key(
        public_exponent=65537, key_size=1024, backend=default_backend()
    )


def gen_passwd():
    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    pw = ""
    for c in range(50):
        pw += random.choice(chars)
    return pw


def generate_csr(key, common_name, email):
    country = "US"
    state = "CA"
    locality = "San Francisco"
    organization_name = "My Org"
    return (
        x509.CertificateSigningRequestBuilder()
        .subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COUNTRY_NAME, country),
                    x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state),
                    x509.NameAttribute(NameOID.LOCALITY_NAME, locality),
                    x509.NameAttribute(
                        NameOID.ORGANIZATION_NAME, organization_name
                    ),
                    x509.NameAttribute(NameOID.COMMON_NAME, common_name),
                    x509.NameAttribute(NameOID.EMAIL_ADDRESS, email),
                ]
            )
        )
        .sign(key, hashes.SHA256(), default_backend())
    )


def gen_pgp_key(email, password, gpg):
    input_data = gpg.gen_key_input(
        name_email=email, passphrase=password, key_type="RSA", key_length=1024
    )
    return gpg.gen_key(input_data)


class User:
    def __init__(self, email, password, key, gpg=None):
        self.gpg = gpg
        self.email = email
        self.password = password
        self.key = key
        self.pgp_key = gen_pgp_key(email, password, gpg)
        self.fingerprint = self.pgp_key.fingerprint
        self.__csrs = []

    @property
    def email(self):
        return self.__email

    @email.setter
    def email(self, email):
        self.__email = email

    @property
    def password(self):
        return self.__password

    @password.setter
    def password(self, password):
        self.__password = password

    @property
    def pgp_key(self):
        return self.__pgp_key

    @pgp_key.setter
    def pgp_key(self, pgp_key):
        self.__pgp_key = pgp_key

    @property
    def csrs(self):
        return self.__csrs

    def gen_csr(self, common_name=None):
        if common_name is None:
            common_name = self.email
        csr = generate_csr(self.key, common_name, self.email)
        self.__csrs.append(csr)
        return csr


class TestCliBase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        TMPDIR_PREFIX = os.environ.get("TMP_DIR") or "/tmp/"
        TMPDIR_PREFIX = os.path.expanduser(TMPDIR_PREFIX)
        VERBOSE_GPG = os.environ.get("DEBUG_GNUPG") or False
        cls.seed_dir = tempfile.TemporaryDirectory(dir=TMPDIR_PREFIX)
        for subpath in ["user", "admin"]:
            os.makedirs(
                "{base}/{subpath}".format(
                    base=cls.seed_dir.name, subpath=subpath
                )
            )
        cls.server_config_dir = tempfile.TemporaryDirectory(dir=TMPDIR_PREFIX)
        cls.docker = docker.from_env()
        cls.ADMIN_GNUPGHOME = tempfile.TemporaryDirectory(dir=TMPDIR_PREFIX)
        cls.admin_gpg = gnupg.GPG(
            gnupghome=cls.ADMIN_GNUPGHOME.name, verbose=VERBOSE_GPG
        )
        cls.admin = User(
            "johndoe@example.com", gen_passwd(), generate_key(), cls.admin_gpg
        )
        file_path = os.path.join(
            cls.seed_dir.name,
            "admin/{}.asc".format(cls.admin.pgp_key.fingerprint),
        )
        with open(file_path, "w") as f:
            f.write(cls.admin_gpg.export_keys(cls.admin.pgp_key.fingerprint))
        cls.USER_GNUPGHOME = tempfile.TemporaryDirectory(dir=TMPDIR_PREFIX)
        cls.user_gpg = gnupg.GPG(
            gnupghome=cls.USER_GNUPGHOME.name, verbose=VERBOSE_GPG
        )
        cls.user = User(
            "johndoe@example.com", gen_passwd(), generate_key(), cls.user_gpg
        )
        file_path = os.path.join(
            cls.seed_dir.name,
            "user/{}.asc".format(cls.user.pgp_key.fingerprint),
        )
        with open(file_path, "w") as f:
            f.write(cls.user_gpg.export_keys(cls.user.pgp_key.fingerprint))
        cls.server_config = ConfigParser()
        cls.server_config_path = os.path.join(
            cls.server_config_dir.name, "config.ini"
        )
        cls.server_config["mtls"] = {
            "min_lifetime": "10",
            "max_lifetime": "1000",
        }
        cls.server_config["ca"] = {
            "key": "secrets/certs/authority/RootCA.key",
            "cert": "secrets/certs/authority/RootCA.pem",
            "issuer": "My Company Name",
            "alternate_name": "*.myname.com",
        }
        cls.server_config["gnupg"] = {
            "user": "secrets/gnupg",
            "admin": "secrets/gnupg_admin",
        }
        cls.server_config["storage"] = {"engine": "sqlite3"}
        cls.server_config["storage.sqlite3"] = {"db_path": "mtls-server.db"}
        with open(cls.server_config_path, "w") as configfile:
            cls.server_config.write(configfile)
        volumes = {}
        volumes[cls.seed_dir.name] = {"bind": "/tmp/seeds", "mode": "ro"}
        volumes[cls.server_config_path] = {
            "bind": "/home/mtls/config.ini",
            "mode": "rw",
        }
        cls.server = cls.docker.containers.run(
            "{image}:{version}".format(
                version=MTLS_SERVER_VERSION, image=MTLS_IMAGE
            ),
            detach=True,
            volumes=volumes,
            remove=True,
            ports={"4000/tcp": 4000},
        )
        while True:
            resp = requests.get("http://localhost:4000/version")
            if resp.status_code == 200:
                break
        cls.HOME = tempfile.TemporaryDirectory(dir=TMPDIR_PREFIX)
        cls.env = {
            "GNUPGHOME": cls.ADMIN_GNUPGHOME.name,
            "HOME": cls.HOME.name,
            "USER": "test",
            "HOST": str(platform.uname()[1]),
        }
        cls.runner = CliRunner(env=cls.env)
        cls.config = ConfigParser()
        cls.config["DEFAULT"] = {
            "name": "John Doe",
            "email": "johndoe@example.com",
            "fingerprint": cls.admin.pgp_key.fingerprint,
            "country": "US",
            "state": "CA",
            "locality": "Mountain View",
            "organization_name": "My Org",
        }
        cls.config["test"] = {"lifetime": 60, "url": "http://localhost:4000"}
        cls.config_path = os.path.join(cls.HOME.name, "config.ini")
        with open(cls.config_path, "w") as configfile:
            cls.config.write(configfile)

    @classmethod
    def tearDownClass(cls):
        cls.server.stop()
        cls.USER_GNUPGHOME.cleanup()
        cls.ADMIN_GNUPGHOME.cleanup()
        cls.seed_dir.cleanup()
        cls.server_config_dir.cleanup()
        cls.HOME.cleanup()


class TestCliAsAdmin(TestCliBase):
    def test_show_help(self):
        result = self.runner.invoke(cli, ["--help"])
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_create_certificate(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
            ],
        )
        self.assertEqual(result.exit_code, 0)

    def test_create_certificate_with_cli_friendly_name_option(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
                "--friendly-name",
                "Foo Bar",
            ],
        )
        self.assertEqual(result.exit_code, 0)
        cert_file_path = os.path.join(self.HOME.name, "test/test.pem")
        with open(cert_file_path, "rb") as cert_file:
            cert = x509.load_pem_x509_certificate(
                cert_file.read(), default_backend()
            )
        self.assertEqual(
            cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value,
            "My Org - Foo Bar",
        )

    def test_create_certificate_with_cli_email_option_no_common_name(self):
        input_email = "test1245566@example.com"
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
                "--user-email",
                input_email,
            ],
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_create_certificate_with_cli_email_option(self):
        input_email = "test1245566@example.com"
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
                "--common-name",
                input_email,
                "--user-email",
                input_email,
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        cert_file_path = os.path.join(self.HOME.name, "test/test.pem")
        with open(cert_file_path, "rb") as cert_file:
            cert = x509.load_pem_x509_certificate(
                cert_file.read(), default_backend()
            )
        email = cert.subject.get_attributes_for_oid(NameOID.EMAIL_ADDRESS)[
            0
        ].value
        self.assertEqual(email, input_email)

    def test_create_certificate_with_cli_output_option(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
                "-o",
                self.HOME.name + "/me.pfx",
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        self.assertTrue(
            os.path.isfile("{}/{}".format(self.HOME.name, "me.pfx"))
        )
        self.assertTrue(
            os.path.isfile("{}/{}".format(self.HOME.name, "me.password.asc"))
        )

    def test_revoke_certificate(self):
        cert_file_path = os.path.join(self.HOME.name, "test/test.pem")
        with open(cert_file_path, "rb") as cert_file:
            cert = x509.load_pem_x509_certificate(
                cert_file.read(), default_backend()
            )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "revoke",
                "--serial-number",
                str(cert.serial_number),
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_add_user_by_fingerprint(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_add_user_by_email(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_remove_user_by_fingerprint(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "remove",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_remove_user_by_email(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_add_admin_by_fingerprint(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_add_admin_by_email(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keys.openpgp.org",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_remove_admin_by_fingerprint(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "remove",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_remove_admin_by_email(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def get_crl_to_output(self):
        result = self.runner.invoke(
            cli, ["-s", "test", "certificate", "crl", "-o"]
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        crl = x509.load_pem_x509_crl(
            data=bytes(result.output, "UTF-8"), backend=default_backend()
        )
        self.assertIsInstance(crl, openssl.x509._CertificateRevocationList)
        self.assertIsInstance(
            crl.get_revoked_certificate_by_serial_number(rev_serial_num),
            openssl.x509._RevokedCertificate,
        )
        self.assertIn(
            "-----BEGIN X509 CRL-----",
            crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
        )
        self.assertIn(
            "-----END X509 CRL-----",
            crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
        )

    def get_crl_to_file(self):
        result = self.runner.invoke(
            cli, ["-s", "test", "certificate", "crl", "-no"]
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        with open("{base}/test/test.crl".format(self.HOME.name), "rb") as crl:
            crl = x509.load_pem_x509_crl(
                data=f.read(), backend=default_backend()
            )
            self.assertIsInstance(crl, openssl.x509._CertificateRevocationList)
            self.assertIsInstance(
                crl.get_revoked_certificate_by_serial_number(rev_serial_num),
                openssl.x509._RevokedCertificate,
            )
            self.assertIn(
                "-----BEGIN X509 CRL-----",
                crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
            )
            self.assertIn(
                "-----END X509 CRL-----",
                crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
            )


class TestCliAsUser(TestCliBase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.env = {
            "GNUPGHOME": cls.USER_GNUPGHOME.name,
            "HOME": cls.HOME.name,
            "USER": "test",
            "HOST": str(platform.uname()[1]),
        }
        cls.runner = CliRunner(env=cls.env)
        cls.config = ConfigParser()
        cls.config["DEFAULT"] = {
            "name": "John Doe",
            "email": "johndoe@example.com",
            "fingerprint": cls.user.pgp_key.fingerprint,
            "country": "US",
            "state": "CA",
            "locality": "Mountain View",
            "organization_name": "My Org",
        }
        cls.config["test"] = {"lifetime": 60, "url": "http://localhost:4000"}
        cls.config_path = os.path.join(cls.HOME.name, "config.ini")
        with open(cls.config_path, "w") as configfile:
            cls.config.write(configfile)

    def test_show_help(self):
        result = self.runner.invoke(cli, ["--help"])
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_create_certificate(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "certificate",
                "create",
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_create_certificate_with_cli_email_option(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.admin.password,
                "certificate",
                "create",
                "--user-email",
                "test1245566@example.com",
            ],
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_revoke_certificate(self):
        cert_file_path = os.path.join(self.HOME.name, "test/test.pem")
        with open(cert_file_path, "rb") as cert_file:
            cert = x509.load_pem_x509_certificate(
                cert_file.read(), default_backend()
            )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "certificate",
                "revoke",
                "--serial-number",
                str(cert.serial_number),
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)

    def test_add_user_by_fingerprint(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_add_user_by_email(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_remove_user_by_fingerprint(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "remove",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_remove_user_by_email(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_add_admin_by_fingerprint(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_add_admin_by_email(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_remove_admin_by_fingerprint(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "remove",
                "--admin",
                "--fingerprint",
                "C92FE5A3FBD58DD3EC5AA26BB10116B8193F2DBD",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def test_remove_admin_by_email(self):
        add_user_result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "user",
                "add",
                "--admin",
                "--keyserver",
                "keyserver.ubuntu.com",
                "--email",
                "danny@drgrovellc.com",
            ],
            input="0",
        )
        self.assertEqual(result.exit_code, 1, msg=result.exc_info)

    def get_crl_to_output(self):
        result = self.runner.invoke(
            cli, ["-s", "test", "certificate", "crl", "-o"]
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        crl = x509.load_pem_x509_crl(
            data=bytes(result.output, "UTF-8"), backend=default_backend()
        )
        self.assertIsInstance(crl, openssl.x509._CertificateRevocationList)
        self.assertIsInstance(
            crl.get_revoked_certificate_by_serial_number(rev_serial_num),
            openssl.x509._RevokedCertificate,
        )
        self.assertIn(
            "-----BEGIN X509 CRL-----",
            crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
        )
        self.assertIn(
            "-----END X509 CRL-----",
            crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
        )

    def get_crl_to_file(self):
        result = self.runner.invoke(
            cli, ["-s", "test", "certificate", "crl", "-no"]
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        with open("{base}/test/test.crl".format(self.HOME.name), "rb") as crl:
            crl = x509.load_pem_x509_crl(
                data=f.read(), backend=default_backend()
            )
            self.assertIsInstance(crl, openssl.x509._CertificateRevocationList)
            self.assertIsInstance(
                crl.get_revoked_certificate_by_serial_number(rev_serial_num),
                openssl.x509._RevokedCertificate,
            )
            self.assertIn(
                "-----BEGIN X509 CRL-----",
                crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
            )
            self.assertIn(
                "-----END X509 CRL-----",
                crl.public_bytes(serialization.Encoding.PEM).decode("UTF-8"),
            )


class TestCliOptions(TestCliBase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.env = {
            "GNUPGHOME": cls.USER_GNUPGHOME.name,
            "HOME": cls.HOME.name,
            "USER": "test",
            "HOST": str(platform.uname()[1]),
        }
        cls.runner = CliRunner(env=cls.env)
        cls.config = ConfigParser()
        cls.config["DEFAULT"] = {
            "name": "John Doe",
            "email": "johndoe@example.com",
            "fingerprint": cls.user.pgp_key.fingerprint,
            "country": "US",
            "state": "CA",
            "locality": "Mountain View",
            "organization_name": "My Org",
        }
        cls.config["test"] = {"lifetime": 60, "url": "http://localhost:4000"}
        cls.config_path = os.path.join(cls.HOME.name, "config.ini")
        with open(cls.config_path, "w") as configfile:
            cls.config.write(configfile)

    def test_add_server(self):
        server_url = "https://certauth.foo.bar"
        result = self.runner.invoke(
            cli,
            ["-c", self.config_path, "server", "add", "foo"],
            input=server_url + "\n",
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        config = ConfigParser()
        config.read(self.config_path)
        self.assertEqual(config.get("foo", "url"), server_url)

    def test_remove_server(self):
        result = self.runner.invoke(
            cli, ["-c", self.config_path, "server", "remove", "foo"]
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        config = ConfigParser()
        config.read(self.config_path)
        self.assertFalse(config.has_section("foo"))

    def test_set_user_config(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "config",
                "organization_name",
                "My New Org",
            ],
        )
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)
        config = ConfigParser()
        config.read(self.config_path)
        self.assertEqual(
            config.get("DEFAULT", "organization_name"), "My New Org"
        )


class TestCliOptionalConfigItems(TestCliBase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.env = {
            "GNUPGHOME": cls.USER_GNUPGHOME.name,
            "HOME": cls.HOME.name,
            "USER": "test",
            "HOST": str(platform.uname()[1]),
        }
        cls.runner = CliRunner(env=cls.env)
        cls.config = ConfigParser()
        cls.config["DEFAULT"] = {
            "name": "John Doe",
            "email": "johndoe@example.com",
            "fingerprint": cls.user.pgp_key.fingerprint,
            "organization_name": "My Org",
        }
        cls.config["test"] = {"lifetime": 60, "url": "http://localhost:4000"}
        cls.config_path = os.path.join(cls.HOME.name, "config.ini")
        with open(cls.config_path, "w") as configfile:
            cls.config.write(configfile)

    def test_create_certificate(self):
        result = self.runner.invoke(
            cli,
            [
                "-c",
                self.config_path,
                "-s",
                "test",
                "--gpg-password",
                self.user.password,
                "certificate",
                "create",
            ],
        )
        if result.exception:
            traceback.print_exception(*result.exc_info)
        self.assertEqual(result.exit_code, 0, msg=result.exc_info)



#!/usr/bin/env python3
import os
import sys
from configparser import ConfigParser
from datetime import datetime

import click

from .mtls import MutualTLS
from . import __version__


HELP_TEXT = (
    "mtls is a PGP Web of Trust based SSL Client Certificate "
    "generation tool based on Googles Beyond Corp Zero Trust "
    "Authentication. Version {}".format(__version__)
)

ALLOWED_KEYS = [
    "name",
    "email",
    "host",
    "fingerprint",
    "country",
    "state",
    "locality",
    "common_name",
    "organization_name",
    "lifetime",
    "url",
]


@click.group(help=HELP_TEXT)
@click.version_option(__version__, message="%(version)s")
@click.option(
    "--server", "-s", type=str, help="Server to run command against."
)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    default=os.path.join(os.environ.get("HOME"), ".config/mtls/config.ini"),
    help="config file. [~/.config/mtls/config.ini]",
)
@click.option("--gpg-password", type=str, hidden=True)
@click.pass_context
def cli(ctx, server, config, gpg_password):
    options = {"config": config, "gpg_password": gpg_password}
    if server is not None:
        ctx.obj = MutualTLS(server, options)
    else:
        ctx.obj = {"config_path": config, "server": server or "DEFAULT"}
    if sys.platform == "win32" or sys.platform == "cygwin":
        click.secho("Your platform is not currently supported", fg="red")


@cli.command(help="Manage config")
@click.argument("key")
@click.argument("value")
@click.pass_context
def config(ctx, key, value):
    AK_MSG = "Your key must be in the allowed keys, available options are: {}"
    # Deal with not actually instanting the MutualTLS class.
    try:
        server = ctx.obj.server or "DEFAULT"
        config_path = ctx.obj.config_file_path
    except Exception as err:
        server = ctx.obj["server"]
        config_path = ctx.obj["config_path"]

    if key not in ALLOWED_KEYS:
        click.secho(AK.MSG.format(",".join(ALLOWED_KEYS)), fg="red")
        sys.exit(1)
    if server == "DEFAULT" and key == "url":
        click.secho(
            "url is not a valid config when no server is set", fg="red"
        )
        sys.exit(1)
    config = ConfigParser()
    config.read(config_path)
    config.set(server, key, value)
    with open(config_path, "w") as config_file:
        config.write(config_file)


@click.group(help="Manage Servers")
@click.pass_context
def server(ctx):
    pass


@server.command("add", help="Add a server")
@click.argument("name")
@click.pass_context
def add_server(ctx, name):
    if name is None or name == "":
        click.secho("Server name cannot be empty", fg="red")
    if " " in name:
        click.secho("Server name cannot have space in it.", fg="red")
        sys.exit(1)
    config_path = ctx.obj["config_path"]
    value = click.prompt(
        "What is the url of the Certificate Authority? (ie. "
        + "https://certauth.example.com): "
    )
    config = ConfigParser()
    config.read(config_path)
    config.add_section(name)
    config.set(name, "url", value)
    with open(config_path, "w") as config_file:
        config.write(config_file)


@server.command("remove", help="Remove a server")
@click.argument("name")
@click.pass_context
def remove_server(ctx, name):
    if name is None or name == "":
        click.secho("Server name cannot be empty", fg="red")
    if " " in name:
        click.secho("Server name cannot have space in it.", fg="red")
        sys.exit(1)
    config_path = ctx.obj["config_path"]
    config = ConfigParser()
    config.read(config_path)
    config.remove_section(name)
    with open(config_path, "w") as config_file:
        config.write(config_file)


@click.group(help="Manage Certificates")
@click.pass_context
def certificate(ctx):
    pass


@certificate.command(
    "create", help="Create a Client Certificate for a given server."
)
@click.option(
    "--output",
    "-o",
    help="""
        Output the pfx file to a location.
        File will not be added to Certificate Store.
    """,
    type=click.Path(exists=False),
)
@click.option(
    "--friendly-name", help="The friendly name of the certificate", type=str
)
@click.option(
    "--user-email", help="The users email for the certificate", type=str
)
@click.option(
    "--organization", "-org", help="The users organization", type=str
)
@click.option(
    "--common-name", "-c", help="The common name for the certificate", type=str
)
@click.pass_context
def create_certificate(
    ctx, output, friendly_name, user_email, organization, common_name
):
    options = {}
    if not isinstance(ctx.obj, MutualTLS):
        click.secho("A server was not provided.", fg="red")
        sys.exit(1)
    if friendly_name:
        options.update(friendly_name=friendly_name)
    if user_email:
        if not common_name:
            click.secho("Must override common name if overriding email")
            sys.exit(1)
        options.update(email=user_email)
    if organization:
        options.update(organization=organization)
    if common_name:
        options.update(common_name=common_name)
    ctx.obj.get_crl(False)
    ctx.obj.set_user_options(options)
    ctx.obj.create_cert(output)


@certificate.command("revoke", help="Revoke a certificate for a given server.")
@click.option("--fingerprint", "-f", default=None, help="User PGP Fingerprint")
@click.option(
    "--serial-number", default=None, help="Serial Number of certificate"
)
@click.option(
    "--name", "-n", default=None, help="The common name on the certificate."
)
@click.pass_context
def revoke_certificate(ctx, fingerprint, serial_number, name):
    if not isinstance(ctx.obj, MutualTLS):
        click.secho("A server was not provided.", fg="red")
        sys.exit(1)
    ctx.obj.revoke_cert(fingerprint, serial_number, name)


@certificate.command("crl", help="Get the CRL for a given server")
@click.option(
    "--output/--no-output",
    "-o/-no",
    is_flag=True,
    default=True,
    help="Output to stdout. Otherwise this will write to "
    + "~/.config/mtls/<server>/crl.pem",
)
@click.pass_context
def get_crl(ctx, output):
    if not isinstance(ctx.obj, MutualTLS):
        click.secho("A server was not provided.", fg="red")
        sys.exit(1)
    ctx.obj.get_crl(output)


@click.group(help="Manage Users")
@click.pass_context
def user(ctx):
    pass


@user.command("add", help="Add a user (Admin Required).")
@click.option(
    "--admin", is_flag=True, default=False, help="Is the user an admin"
)
@click.option("--fingerprint", "-f", default=None, help="User PGP Fingerprint")
@click.option(
    "--email",
    "-e",
    default=None,
    help="User email. This will grab the users fingerprint from your local "
    + "trust store",
)
@click.option(
    "--keyserver",
    default=None,
    help="Keyserver for searching by email. Defaults to pgp.mit.edu",
)
@click.pass_context
def add_user(ctx, admin, fingerprint, email, keyserver):
    if not isinstance(ctx.obj, MutualTLS):
        click.secho("A server was not provided.", fg="red")
        sys.exit(1)
    if fingerprint is None and email is None:
        click.echo("A fingerprint or email must be provided")
        sys.exit(1)
    if email is not None:
        fingerprint = handle_email(ctx, email, keyserver)
    ctx.obj.add_user(fingerprint, admin)


@user.command("remove", help="Remove a user (Admin Required).")
@click.option(
    "--admin", is_flag=True, default=False, help="Is the user an admin"
)
@click.option("--fingerprint", "-f", default=None, help="User PGP Fingerprint")
@click.option(
    "--email",
    "-e",
    default=None,
    help="User email. This will grab the users fingerprint from your local "
    + "trust store",
)
@click.option(
    "--keyserver",
    default=None,
    help="Keyserver for searching by email. Defaults to pgp.mit.edu",
)
@click.pass_context
def remove_user(ctx, admin, fingerprint, email, keyserver):
    if not isinstance(ctx.obj, MutualTLS):
        click.secho("A server was not provided.", fg="red")
        sys.exit(1)
    if fingerprint is None and email is None:
        click.echo("A fingerprint or email must be provided")
        sys.exit(1)
    if email is not None:
        fingerprint = handle_email(ctx, email, keyserver)

    ctx.obj.remove_user(fingerprint, admin)


def handle_email(ctx, email, keyserver=None):
    if keyserver:
        search_res = ctx.obj.gpg.search_keys(email, keyserver=keyserver)
    else:
        search_res = ctx.obj.gpg.search_keys(email)
    now = str(int(datetime.now().timestamp()))
    non_expired = []
    for res in search_res:
        if res["expires"] == "" or res["expires"] > now:
            non_expired.append(res)
    if len(non_expired) == 0:
        click.secho("A fingerprint with the key could not be found.")
        sys.exit(1)
    if len(non_expired) == 1:
        return non_expired[0]["keyid"]
    for idx, res in enumerate(non_expired):
        click.echo(
            "{idx}) {fingerprint} {uid} - Created: {created}".format(
                idx=idx,
                fingerprint=res["keyid"],
                uid=res["uids"][0],
                created=datetime.utcfromtimestamp(int(res["date"])).strftime(
                    "%m/%d/%Y %H:%M:%S"
                ),
            )
        )
    num = len(non_expired)
    try:
        value = int(input("Please select a key to add: "))
        if value > num:
            click.secho("Invalid number, exiting")
            sys.exit(1)
    except Exception:
        click.secho("Invalid number, exiting")
        sys.exit(1)
    return non_expired[value]["keyid"]


# Bind the subcommands to the cli
cli.add_command(certificate)
cli.add_command(user)
cli.add_command(server)



from setuptools import setup, find_packages

desc = "A short-lived certificate tool based on the Zero Trust network mode"

setup(
    name="mtls",
    author="Danny Grove <danny@drgrovellc.com>",
    url="https://github.com/drGrove/mtls-cli",
    description=desc,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    project_urls={
        "Homepage": "https://github.com/drGrove/mtls-cli",
        "Source": "https://github.com/drGrove/mtls-cli",
        "Tracker": "https://github.com/drGrove/mtls-cli/issues",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Security :: Cryptography",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires='>=3.6',
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    packages=find_packages(exclude=["test"]),
    package_data={"mtls": ["share/*"]},
    entry_points={
        "console_scripts": [
            "mtls = mtls.cli:cli"
        ]
    },
    install_requires=[
        "cryptography",
        "python-gnupg",
        "urllib3",
        "requests",
        "click",
        "pyOpenSSL",
    ],
    zip_safe=True
)



import os
import sys

from pkg_resources import get_distribution, DistributionNotFound

__author__ = "Danny Grove <danny@drgrovellc.com>"

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    __version__ = "dev"

# Allows "import mtls" and "from mtls import <name>".
sys.path.extend([os.path.join(os.path.dirname(__file__), "..")])

from .cli import cli  # noqa
from .mtls import MutualTLS  # noqa
